/* eslint-disable no-unused-expressions */
/* eslint-disable react-hooks/exhaustive-deps */
import React, { useState, useRef, useEffect } from 'react';
import PropTypes, { element } from 'prop-types';
import loadable from '@loadable/component';

import {
  updateSlider,
  swipeStart,
  swipeMove,
  swipeStop,
  conditionalClick,
  debounceAnimationFrame,
} from '~helpers/slider';

import styles from './slider.module.scss';
import Arrow from '~atoms/carousel/arrow';

function Slider({
  gtmClass = '',
  useArrows = true,
  usePagination = true,
  arrows = {
    prev: <Arrow direction="left" gtmClass={gtmClass} />,
    next: <Arrow direction="right" gtmClass={gtmClass} />,
  },
  settings = {
    swipeEndThreshold: 0.25,
    transitionTime: 400,
    numberOfSlides: 1,
    sliderGap: 0,
  },
  className = '',
  children,
}) {
  const [state, setState] = useState({
    action: '',
    slideIndex: 0,
    swipeOffset: 0,
    swipeDirection: null,
    swipeStatus: 'stop',
    length: children.length,
  });

  const sliderContainerRef = useRef(null);
  const sliderContentRef = useRef(null);
  const sliderIsActive = useRef(false);
  const sliderConditionsRef = useRef({
    autoSlideWidth: settings.numberOfSlides === 0,
    slideWidth: 0,
    totalSlideWidth: 0,
    containerWidth: 0,
  });

  let style = {};

  // Update Slider on Swipe
  const updateOnSwipe = (e, status, swipeStartPoint) => {
    let offset = 0;
    let direction = 0;

    if (e.touches) {
      if (e.touches.length > 0) offset = e.touches[0].clientX;
      else offset = e.changedTouches[0].clientX;
    } else {
      offset = e.clientX;
    }

    offset -= swipeStartPoint;
    direction = Math.sign(offset);
    setState({
      ...state,
      swipeDirection: direction,
      swipeOffset: offset,
      swipeStatus: status,
      action: 'swipe',
    });
  };

  const paginationHandler = (index) => {
    setState({
      ...state,
      action: 'nav',
      slideIndex: index * -1,
    });
  };

  const arrowHandler = (action) => {
    let index = 0;

    if (action === 'prev') {
      index = Math.abs(state.slideIndex) > 0 ? state.slideIndex + 1 : 0;
    } else if (action === 'next') {
      index =
        Math.abs(state.slideIndex) < state.length - 1 ? state.slideIndex - 1 : state.slideIndex;
    }

    setState({
      ...state,
      action: `nav`,
      slideIndex: index,
    });
  };

  [style, sliderConditionsRef.current] = updateSlider(
    style,
    state,
    sliderContentRef,
    sliderContainerRef,
    settings,
    sliderConditionsRef
  );

  let throttlePause;
  const onWheel = (e) => {
    e.preventDefault();
    if (useArrows || usePagination) {
      throttlePause ? null : (throttlePause = true);
      setTimeout(() => {
        e.deltaY > 0 ? arrowHandler('next') : arrowHandler('prev');
        throttlePause = false;
      }, 300);
    } else {
      sliderContainerRef.current.scrollTo({
        top: 0,
        left: sliderContainerRef.current.scrollLeft + e.deltaY,
      });
    }
  };

  useEffect(() => {
    // Keep slider node reference for multiple carousels in an instance
    const sliderContentNode = sliderContentRef.current;

    sliderContentNode.style.setProperty('width', style.width);
    sliderContentNode.style.setProperty('transition', style.transition);
    sliderContentNode.style.setProperty('transform', style.transform);
    sliderContentNode.style.setProperty('grid-template-columns', style.gridTemplateColumns);
    sliderContentNode.style.setProperty('gap', style.gridGap);

    const mousedown = (e) => {
      sliderIsActive.current = true;
      swipeStart(e, updateOnSwipe);
    };

    const mousemove = (e) => {
      if (sliderIsActive.current) {
        swipeMove(e, updateOnSwipe);
      }
    };

    const mouseup = (e) => {
      if (sliderIsActive.current) {
        sliderIsActive.current = false;
        swipeStop(e, updateOnSwipe);
      }
    };

    const click = (e) => {
      conditionalClick(e);
    };

    // Debounce Swipe updates functions
    const mouseMoveDebounce = debounceAnimationFrame(mousemove);
    const mouseUpDebounce = debounceAnimationFrame(mouseup);

    sliderContentNode.addEventListener('click', click);
    sliderContentNode.addEventListener('mousedown', mousedown);
    document.addEventListener('mousemove', mouseMoveDebounce);
    document.addEventListener('mouseup', mouseUpDebounce);
    sliderContentNode.addEventListener('touchstart', mousedown, { passive: true });
    sliderContentNode.addEventListener('touchmove', mouseMoveDebounce, { passive: true });
    sliderContentNode.addEventListener('touchend', mouseUpDebounce);
    sliderContainerRef.current.addEventListener('wheel', onWheel, { passive: true });

    return () => {
      sliderContentNode?.removeEventListener('click', click);
      sliderContentNode?.removeEventListener('mousedown', mousedown);
      document.removeEventListener('mousemove', mouseMoveDebounce);
      document.removeEventListener('mouseup', mouseUpDebounce);
      sliderContentNode?.removeEventListener('touchstart', mousedown, { passive: true });
      sliderContentNode?.removeEventListener('touchmove', mouseMoveDebounce, { passive: true });
      sliderContentNode?.removeEventListener('touchend', mouseUpDebounce);
      sliderContainerRef?.current?.removeEventListener('wheel', onWheel, { passive: true });
    };
  });

  const Pagination = usePagination
    ? loadable(() => import('~atoms/carousel/pagination-item'))
    : null;

  return (
    state.length > 0 && (
      <div className={`${className || ''} ${styles.slider || ''}`} ref={sliderContainerRef}>
        {useArrows &&
          React.cloneElement(arrows.prev, {
            onClick: () => arrowHandler('prev'),
            onKeyDown: () => arrowHandler('prev'),
            disabled: state.slideIndex === 0 && true,
          })}
        {useArrows &&
          React.cloneElement(arrows.next, {
            onClick: () => arrowHandler('next'),
            onKeyDown: () => arrowHandler('next'),
            disabled: Math.abs(state.slideIndex) === state.length - 1 && true,
          })}
        <div className={styles.sliderContent || ''} style={style} ref={sliderContentRef}>
          {children.map((slide, index) => (
            <div
              // eslint-disable-next-line react/no-array-index-key
              key={index}
              className={`${styles.sliderItem || ''}
              ${settings.numberOfSlides !== 0 ? styles.sliderHeight : ''}
              ${
                settings.numberOfSlides !== 0 && Math.abs(state.slideIndex) === index
                  ? styles.active
                  : ''
              }`}
            >
              {React.cloneElement(slide, {
                draggable: 'true',
                onDragStart: (e) => e.preventDefault(),
              })}
            </div>
          ))}
        </div>
        {Pagination && (
          <Pagination
            type="dots"
            paginationHandler={paginationHandler}
            slideIndex={state.slideIndex}
            totalSlides={state.length}
            gtmClass={gtmClass}
          />
        )}
      </div>
    )
  );
}

Slider.propTypes = {
  useArrows: PropTypes.bool,
  usePagination: PropTypes.bool,
  arrows: PropTypes.shape({
    prev: PropTypes.element,
    next: PropTypes.element,
  }),
  settings: PropTypes.shape({
    swipeEndThreshold: PropTypes.number,
    transitionTime: PropTypes.number,
    numberOfSlides: PropTypes.number,
  }),
  className: PropTypes.string,
  children: PropTypes.arrayOf(element).isRequired,
  gtmClass: PropTypes.string,
};

export default Slider;
